/*-
 * Copyright 2022 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <string.h>	// memset() for INIT_OBJ()
#include <stdlib.h>	// strtoul()

#include <cache/cache.h>
#include <cache/cache_filter.h>
#include <vcl.h>

#include "vcc_zipflow_if.h"

#include "zlib.h"
#include "zipflow.h"

static void __attribute__((constructor))
assert_zlib(void)
{
	assert(Z_DEFAULT_COMPRESSION == -1);
}

// https://github.com/varnishcache/varnish-cache/pull/3815
#ifndef WS_TASK_ALLOC_OBJ
#define WS_TASK_ALLOC_OBJ(ctx, ptr, magic) do {			\
	ptr = WS_Alloc((ctx)->ws, sizeof *(ptr));		\
	if ((ptr) == NULL)					\
		VRT_fail(ctx, "Out of workspace for " #magic);	\
	else							\
		INIT_OBJ(ptr, magic);				\
} while(0)
#endif

static char default_level = Z_DEFAULT_COMPRESSION;

static const void *zipflow_request_priv = &zipflow_request_priv;
static const void *zipflow_top_priv = &zipflow_top_priv;

struct zipflow_top;
struct zipflow_request {
	unsigned			magic;
#define ZIPFLOW_REQUEST_MAGIC		0xaa175160
	unsigned			bundle:1;
	char				level;
	VCL_STRING			url;
	VCL_STRING			name;
	unsigned			mode;
	uint32_t			atime, mtime;
	VSTAILQ_ENTRY(zipflow_request)	list;
	struct zipflow_top		*top;
};

VSTAILQ_HEAD(zipflow_head, zipflow_request);

struct zipflow_top {
	unsigned		magic;
#define ZIPFLOW_TOP_MAGIC	0x5743145e
	struct zipflow_head	head;
	ZIP			*zip;
};

static struct zipflow_top *
get_zipflow_top(VRT_CTX)
{
	struct vmod_priv *task_priv;
	struct zipflow_top *zft;

	task_priv = VRT_priv_task(ctx, zipflow_top_priv);
	if (task_priv == NULL) {
		VRT_fail(ctx, "Can not get PRIV_TASK");
		return (NULL);
	}

	if (task_priv->priv != NULL) {
		CAST_OBJ_NOTNULL(zft, task_priv->priv, ZIPFLOW_TOP_MAGIC);
		return (zft);
	}

	WS_TASK_ALLOC_OBJ(ctx, zft, ZIPFLOW_TOP_MAGIC);
	if (zft == NULL)
		return (NULL);

	VSTAILQ_INIT(&zft->head);
	return (zft);
}

static struct vdp vdp_zipflow;

static struct zipflow_request *
new_zipflow_request(VRT_CTX, struct zipflow_top *zft)
{
	struct zipflow_request *zfr;

	WS_TASK_ALLOC_OBJ(ctx, zfr, ZIPFLOW_REQUEST_MAGIC);
	if (zfr == NULL)
		return (NULL);

	VSTAILQ_INSERT_TAIL(&zft->head, zfr, list);

	zfr->bundle = 1;
	zfr->level = default_level;
	zfr->name = "unnamed_file";
	zfr->mode = 0644;
	zfr->atime = ctx->now;
	zfr->mtime = ctx->now;
	zfr->top = zft;
	return (zfr);
}

static struct zipflow_request *
get_zipflow_request(VRT_CTX)
{
	struct vmod_priv *task_priv;
	struct zipflow_request *zfr;
	struct zipflow_top *zft;

	task_priv = VRT_priv_task(ctx, zipflow_request_priv);
	if (task_priv == NULL) {
		VRT_fail(ctx, "Can not get PRIV_TASK");
		return (NULL);
	}

	if (task_priv->priv != NULL) {
		CAST_OBJ_NOTNULL(zfr, task_priv->priv, ZIPFLOW_REQUEST_MAGIC);
		return (zfr);
	}

	zft = get_zipflow_top(ctx);
	if (zft == NULL)
		return (NULL);

	zfr = new_zipflow_request(ctx, zft);
	if (zfr == NULL)
		return (NULL);

	task_priv->priv = zfr;
	return (zfr);
}

/*
 * ============================================================
 * VCL interface
 */

VCL_VOID
vmod_subreq(VRT_CTX, VCL_STRING url)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(ctx->method & (VCL_MET_TASK_C));

	zft = get_zipflow_top(ctx);
	zfr = new_zipflow_request(ctx, zft);

	zfr->url = url;
}

VCL_BOOL
vmod_is_subreq(VRT_CTX)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	// XXX TODO
	return (1);
}

VCL_VOID
vmod_bundle(VRT_CTX, VCL_BOOL toggle)
{
	struct zipflow_request *zfr;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	zfr = get_zipflow_request(ctx);
	if (zfr == NULL)
		return;

	zfr->bundle = toggle ? 1 : 0;
}

VCL_VOID
vmod_set_level(VRT_CTX, VCL_INT level)
{
	struct zipflow_request *zfr;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (level < -1 || level > 9) {
		VRT_fail(ctx, "invalid level %ld", level);
		return;
	}

	if (ctx->method == VCL_MET_INIT) {
		default_level = level;
		return;
	}
	AZ(ctx->method & VCL_MET_TASK_H);

	AN(ctx->method & VCL_MET_TASK_C);
	zfr = get_zipflow_request(ctx);
	if (zfr == NULL)
		return;
	zfr->level = (char)level;
}

VCL_VOID
vmod_meta(VRT_CTX, VCL_STRING name, VCL_STRING mode,
    VCL_TIME atime, VCL_TIME mtime)
{
	unsigned long u;
	struct zipflow_request *zfr = get_zipflow_request(ctx);

	if (zfr == NULL)
		return;

	u = strtoul(mode, NULL, 8);
	if (u == ULONG_MAX) {
		VRT_fail(ctx, ".meta() error converting mode: %d (%s)",
		    errno, VAS_errtxt(errno));
		return;
	}
	if (atime == 0)
		atime = ctx->now;
	if (mtime == 0)
		mtime = ctx->now;

	zfr->name = name;
	zfr->mode = u;
	zfr->atime = atime;
	zfr->mtime = mtime;
}

int
vmod_event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	(void) priv;

	switch(e) {
	case VCL_EVENT_LOAD:
		if (VRT_AddFilter(ctx, NULL, &vdp_zipflow) != NULL)
			return (1);
		break;
	case VCL_EVENT_DISCARD:
		VRT_RemoveFilter(ctx, NULL, &vdp_zipflow);
		break;
	case VCL_EVENT_WARM:
	case VCL_EVENT_COLD:
		break;
	default:
		WRONG("illegal event enum");
	}
	return (0);
}

/*
 * ============================================================
 * VDP
 */


static int
vdp_zipflow_put(void *priv, void const *ptr, size_t len)
{

	if (ptr == NULL || len == 0)
		return (0);

	return (VDP_bytes(priv, VDP_FLUSH, ptr, len));
}

static void
vdp_zipflow_log(void *vsl, char *msg)
{

	AN(vsl);
	AN(msg);
	VSLb(vsl, SLT_Error, "zipflow: %s", msg);
	free(msg);
}

static int
vdp_zipflow_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, struct objcore *oc)
{
	struct zipflow_request *zfr = get_zipflow_request(ctx);
	struct zipflow_top *zft;
	struct req *req;

	if (zfr == NULL)
		return (1);

	AN(zfr->name);

	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);

	assert(zfr == VSTAILQ_FIRST(&zft->head));
	VSTAILQ_REMOVE_HEAD(&zft->head, list);

	AZ(zft->zip);
	zft->zip = zip_pipe(vdc, vdp_zipflow_put, zfr->level);
	if (zft->zip == NULL) {
		VSLb(vdc->vsl, SLT_Error, "zipflow: zip_pipe failed");
		return (1);
	}
	AZ(zip_log(zft->zip, vdc->vsl, vdp_zipflow_log));
	if (zfr->bundle) {
		AZ(zip_meta(zft->zip, zfr->name, 3, zfr->mode,
		    zfr->atime, zfr->mtime));
	}

	AN(priv);
	AZ(*priv);
	*priv = zfr;

	req = vdc->req;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	RFC2616_Weaken_Etag(req->resp);
	if (req->resp_len != 0)
		req->resp_len = -1;

	return (0);
}

static int
vdp_zipflow_fini(struct vdp_ctx *vdc, void **priv)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	int r;

	AN(priv);
	zfr = *priv;
	*priv = NULL;
	if (zfr != NULL) {
		CHECK_OBJ(zfr, ZIPFLOW_REQUEST_MAGIC);
		zft = zfr->top;
		CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);
		r = zip_close(zft->zip);
		if (r)
			VSLb(vdc->vsl, SLT_Error, "zip_close returned %d", r);
		memset(zfr, 0, sizeof *zfr);
	}
	return (0);
}

static int
vdp_zipflow_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
    const void *ptr, ssize_t len)
{
	struct zipflow_request *zfr;
	struct zipflow_top *zft;
	int r;

	AN(priv);
	CAST_OBJ_NOTNULL(zfr, *priv, ZIPFLOW_REQUEST_MAGIC);
	zft = zfr->top;
	CHECK_OBJ_NOTNULL(zft, ZIPFLOW_TOP_MAGIC);
	if (zfr->bundle == 0) {
		(void) zip_close(zft->zip);
		return (1);
	}
	r = zip_data(zft->zip, ptr, len, act == VDP_END);
	if (r) {
		VSLb(vdc->vsl, SLT_Error, "zip_data returned %d", r);
		return (-1);
	}

	if (act != VDP_END)
		return (0);

	*priv = NULL;
	r = zip_close(zft->zip);
	if (r)
		VSLb(vdc->vsl, SLT_Error, "zip_close returned %d", r);
	memset(zfr, 0, sizeof *zfr);
	return (VDP_bytes(vdc, VDP_END, NULL, 0));
}

static struct vdp vdp_zipflow = {
	.name = "zipflow",
	.init = vdp_zipflow_init,
	.bytes = vdp_zipflow_bytes,
	.fini = vdp_zipflow_fini
};
